# set on model, represet the define
# model query and mutate name when querying
# and mutating the model
QL_QUERY_NAME_ATTR = "__ql_query_name__"
QL_MUTATE_NAME_ATTR = "__ql_mutate_name__"
QL_TYPENAME_ATTR = "__ql_typename__"

# represent the model implementions
QL_IMPLEMENTS_ATTR = "__ql_implements__"

QL_INSTANTIATE = "__ql_instantiate__"

# returns namedtuple for accessing
# model fields with dot
QL_MUTABLE_FIELDS_NT_ATTR = "__ql_mutate_fields_nt__"
QL_QUERYABLE_FIELDS_NT_ATTR = "__ql_query_fields_nt__"
